# Generated by Django 3.2.13 on 2022-06-09 11:26

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("caluma_analytics", "0003_starting_objects"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="historicalanalyticsfield",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical analytics field",
                "verbose_name_plural": "historical analytics fields",
            },
        ),
        migrations.AlterModelOptions(
            name="historicalanalyticstable",
            options={
                "get_latest_by": ("history_date", "history_id"),
                "ordering": ("-history_date", "-history_id"),
                "verbose_name": "historical analytics table",
                "verbose_name_plural": "historical analytics tables",
            },
        ),
        migrations.AlterField(
            model_name="historicalanalyticsfield",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
        migrations.AlterField(
            model_name="historicalanalyticstable",
            name="history_date",
            field=models.DateTimeField(db_index=True),
        ),
    ]



from graphlib import TopologicalSorter
from logging import getLogger
from typing import Optional

from django.db import transaction
from django.db.models import Q
from rest_framework.exceptions import ValidationError

from caluma.caluma_core.exceptions import ConfigurationError
from caluma.caluma_core.models import BaseModel
from caluma.caluma_core.relay import extract_global_id
from caluma.caluma_form import models, validators
from caluma.caluma_form.utils import recalculate_field
from caluma.caluma_user.models import BaseUser
from caluma.utils import update_model

log = getLogger(__name__)


class BaseLogic:
    @staticmethod
    @transaction.atomic
    def create(model: BaseModel, validated_data, user: Optional[BaseUser] = None):
        if user:
            validated_data["created_by_user"] = user.username if user else None
            validated_data["created_by_group"] = user.group if user else None
            validated_data["modified_by_user"] = user.username if user else None
            validated_data["modified_by_group"] = user.group if user else None

        return model.objects.create(**validated_data)

    @staticmethod
    @transaction.atomic
    def update(instance: BaseModel, validated_data, user: Optional[BaseUser] = None):
        if user:
            validated_data["modified_by_user"] = user.username
            validated_data["modified_by_group"] = user.group

        update_model(instance, validated_data)
        instance.refresh_from_db()

        return instance


class SaveAnswerLogic:
    @classmethod
    def get_new_answer(cls, data, user, answer):
        validated_data = cls.pre_save(
            cls.validate_for_save(data, user, answer=answer, origin=True)
        )

        files = validated_data.pop("files", None)
        if answer is None:
            answer = cls.create(validated_data, user)
        else:
            answer = cls.update(answer, validated_data, user)

        # FILES type needs a bit more work due to the reverse
        # relation of file -> answer
        if answer.question.type == models.Question.TYPE_FILES:
            if files is None:
                raise ValidationError("Files input must be a list")
            cls.update_answer_files(answer, files)

        return cls.post_save(answer)

    @classmethod
    def update_answer_files(cls, answer: models.Answer, files: list):
        """Update the files of a "FILES" answer.

        The files parameter is expected to be a list of dicts, where
        each entry has a "name" (the file name), and optionally an "id"
        for the case when the given file already exists.
        """
        if not files:
            files = []

        updated = []

        for file_ in files:
            file_id = extract_global_id(file_["id"]) if "id" in file_ else None
            file_name = file_["name"]
            file_model = (
                answer.files.filter(pk=file_id).first()
                if file_id
                else models.File(answer=answer, name=file_name)
            )
            if not file_model:
                # Client wants to update a file that doesn't exist anymore.
                # Reject call - this is an inconsistency as either
                # the file was never created, or was deleted in the mean
                # time.
                raise ValidationError(
                    f"File with id={file_id} for given answer not found"
                )

            if file_model.name != file_name and file_id:
                # Renaming existing file
                file_model.rename(file_name)
                file_model.save()
            elif not file_id:
                # New file
                file_model.save()
            updated.append(file_model.pk)

        for file in answer.files.exclude(pk__in=updated):
            file.delete()

    @staticmethod
    def validate_for_save(
        data: dict,
        user: BaseUser,
        answer: models.Answer = None,
        origin: bool = False,
        data_source_context: dict = None,
    ) -> dict:
        question = data["question"]

        # Save empty answer by omitting the value property since graphql-core
        # doesn't allow `null` literals yet
        if "value" not in data:
            data["value"] = None

        # make data from python api similar to graphql api
        if question.type == models.Question.TYPE_TABLE and not data.get("documents"):
            data["documents"] = (
                models.Document.objects.filter(pk__in=data["value"])
                if data["value"]
                else models.Document.objects.none()
            )
            del data["value"]
        elif question.type == models.Question.TYPE_FILES and not data.get("files"):
            data["files"] = data["value"]
            del data["value"]
        elif question.type == models.Question.TYPE_DATE and not data.get("date"):
            data["date"] = data["value"]
            del data["value"]

        validators.AnswerValidator().validate(
            **data,
            user=user,
            instance=answer,
            origin=origin,
            data_source_context=data_source_context,
        )

        return data

    @staticmethod
    def pre_save(validated_data: dict) -> dict:
        # TODO emit events
        return validated_data

    @staticmethod
    def post_save(answer: models.Answer) -> models.Answer:
        # TODO emit events
        return answer

    @staticmethod
    def recalculate_dependents(answer, update_info):
        """Update the dependent calc answers when this given answer has changed.

        The update_info is used in the context of table questions, when we need
        to know in which rows to look for calc questions (as they may depend on
        out-of-row questions)
        """
        is_table = update_info and answer.question.type == models.Question.TYPE_TABLE
        if not is_table and not answer.question.calc_dependents:
            return
        if not answer.document:
            # Default answer
            return
        log.debug("update_calc_dependents(%s)", answer)

        struc = validators.DocumentValidator().get_validation_context(
            answer.document.family
        )
        field = struc.find_field_by_answer(answer)
        if not field:  # pragma: no cover
            # not covering this, because it's a developer error
            raise ConfigurationError(
                "Saved an answer in a form structure where it doesn't belong: "
                f"question={answer.question_id}, root form={struc.get_root().get_form().slug}",
            )
        if is_table:
            # We treat all children of the table as "changed"
            # Find all children, and if any are of type calc (and are in the "created"
            # update info bit), recalculate them.
            for child in field.get_all_fields():
                if (
                    child.question.type == models.Question.TYPE_CALCULATED_FLOAT
                    and child.parent._document.pk in update_info["created"]
                ):
                    recalculate_field(child)

        for dep_slug in field.question.calc_dependents or []:
            # ... maybe this is enough? Cause this will find the closest "match",
            # going only to the outer context from a table if the field is not found
            # inside of it
            for dep_field in struc.find_all_fields_by_slug(dep_slug):
                # Need to iterate, because a calc question could reside *inside*
                # a table row, so there could be multiple of them, all needing to
                # be updated because of "our" change

                log.debug(
                    "update_calc_dependents(%s): updating question %s",
                    answer,
                    dep_field.question.pk,
                )
                recalculate_field(dep_field)

    @classmethod
    @transaction.atomic
    def create(
        cls, validated_data: dict, user: Optional[BaseUser] = None
    ) -> models.Answer:
        if validated_data["question"].type == models.Question.TYPE_TABLE:
            documents = validated_data.pop("documents")

        files = validated_data.pop("files", None)
        answer = BaseLogic.create(models.Answer, validated_data, user)

        if validated_data["question"].type == models.Question.TYPE_FILES:
            cls.update_answer_files(answer, files)

        update_info = None
        if answer.question.type == models.Question.TYPE_TABLE:
            update_info = answer.create_answer_documents(documents)

        cls.recalculate_dependents(answer, update_info)

        return answer

    @classmethod
    @transaction.atomic
    def update(cls, answer, validated_data, user: Optional[BaseUser] = None):
        if answer.question.type == models.Question.TYPE_TABLE:
            documents = validated_data.pop("documents")
            answer.unlink_unused_rows(docs_to_keep=documents)

        if answer.question.type == models.Question.TYPE_FILES:
            cls.update_answer_files(answer, validated_data.pop("files", None))

        BaseLogic.update(answer, validated_data, user)

        update_info = None
        if answer.question.type == models.Question.TYPE_TABLE:
            update_info = answer.create_answer_documents(documents)

        cls.recalculate_dependents(answer, update_info)

        answer.refresh_from_db()
        return answer


class SaveDefaultAnswerLogic(SaveAnswerLogic):
    @staticmethod
    def validate_for_save(
        data: dict, user: BaseUser, answer: models.Answer = None, origin: bool = False
    ) -> dict:
        if data["question"].type in [
            models.Question.TYPE_FILES,
            models.Question.TYPE_STATIC,
            models.Question.TYPE_DYNAMIC_CHOICE,
            models.Question.TYPE_DYNAMIC_MULTIPLE_CHOICE,
            models.Question.TYPE_CALCULATED_FLOAT,
            models.Question.TYPE_ACTION_BUTTON,
        ]:
            raise ValidationError(
                f'Can\'t save default_answer for question of type "{data["question"].type}"'
            )

        data["document"] = None  # send None as document for validation
        return SaveAnswerLogic.validate_for_save(data, user, answer, origin)

    @staticmethod
    @transaction.atomic
    def create(validated_data: dict, user: Optional[BaseUser] = None) -> models.Answer:
        answer = SaveAnswerLogic.create(validated_data, user)
        answer.question.default_answer = answer
        answer.question.save()

        return answer

    @staticmethod
    @transaction.atomic
    def update(
        answer, validated_data, user: Optional[BaseUser] = None
    ) -> models.Answer:
        return SaveAnswerLogic.update(answer, validated_data, user)


class SaveDocumentLogic:
    @staticmethod
    def _set_default_answers_for_form(form, document, user):
        """
        Set answers for questions with a default_answer.

        This method sets the default_answers on a new document. It recusively travels
        through the form and it's subforms, looking for questions that have a
        default_answer.
        For adding the answers, it does one query per form and one query per
        row_document.
        """
        for question in form.questions.filter(
            Q(default_answer__isnull=False) | Q(type=models.Question.TYPE_FORM)
        ).iterator():
            if question.type == models.Question.TYPE_FORM:
                if question.sub_form is not None:
                    SaveDocumentLogic._set_default_answers_for_form(
                        question.sub_form, document, user
                    )
                continue

            question.default_answer.copy(
                to_document=document, document_family=document.family, user=user
            )

        return document

    @staticmethod
    @transaction.atomic
    def create(
        validated_data: dict, user: Optional[BaseUser] = None
    ) -> models.Document:
        # When creating a new document, we want to defer the calculation of calculated
        # answers until after all the default answers are created. Otherwise, this would
        # start a huge matrix of recalculations.
        validated_data.setdefault("meta", {})
        validated_data["meta"]["_defer_calculation"] = True

        document = BaseLogic.create(models.Document, validated_data, user=user)

        document = SaveDocumentLogic._set_default_answers_for_form(
            document.form, document, user
        )

        # Now, all default answers are created. We remove the _defer_calculation-key
        # from the meta and manually trigger the calculation.
        document.meta.pop("_defer_calculation", None)
        document.save()

        SaveDocumentLogic._initialize_calculated_answers(document)

        return document

    @staticmethod
    def _initialize_calculated_answers(document):
        """
        Initialize all calculated questions in the document.

        In order to do this efficiently, we get all calculated questions with
        their dependents, sort them topoligically, and then update their answer.
        """
        struc = validators.DocumentValidator().get_validation_context(document.family)

        calculated_fields = (
            field
            for field in struc.get_all_fields()
            if field.question.type == models.Question.TYPE_CALCULATED_FLOAT
        )

        adjacency_list = {
            field.slug(): field.question.calc_dependents for field in calculated_fields
        }
        ts = TopologicalSorter(adjacency_list)
        # TopologicalSorter expects the adjacency_list the "other way around", i.e.
        # for every node the incoming nodes should be given. To account for this, we
        # just reverse the resulting order.
        sorted_question_slugs = list(reversed(list(ts.static_order())))

        for slug in sorted_question_slugs:
            print("question", slug)
            for field in struc.find_all_fields_by_slug(slug):
                recalculate_field(field, update_recursively=False)

        return document

    @staticmethod
    @transaction.atomic
    def update(document, validated_data, user: Optional[BaseUser] = None):
        BaseLogic.update(document, validated_data, user)


class CopyFormLogic:
    @staticmethod
    def copy(validated_data: dict, user: Optional[BaseUser] = None):
        source = validated_data["source"]
        validated_data["meta"] = dict(source.meta)
        form = BaseLogic.create(models.Form, validated_data, user=user)

        for form_question in models.FormQuestion.objects.filter(form=source):
            BaseLogic.create(
                models.FormQuestion,
                {
                    "form": form,
                    "sort": form_question.sort,
                    "question": form_question.question,
                },
                user,
            )

        return form



# Some common exceptions


class ConfigurationError(Exception):
    """Invalid configuration detected.

    Use this exception type if a configuration does not make
    sense or is generally un-processable.

    For example: circular dependencies in JEXL expressions,
    invalid form hierarchies etc
    """


class QuestionMissing(Exception):
    pass



# Generated by Django 3.2.16 on 2023-01-31 09:25

import localized_fields.fields.field
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("caluma_analytics", "0006_rename_form_slug_fields"),
    ]

    operations = [
        migrations.AddField(
            model_name="analyticstable",
            name="description",
            field=localized_fields.fields.field.LocalizedField(
                blank=True, null=True, required=[]
            ),
        ),
        migrations.AddField(
            model_name="historicalanalyticstable",
            name="description",
            field=localized_fields.fields.field.LocalizedField(
                blank=True, null=True, required=[]
            ),
        ),
    ]



# Generated by Django 4.2.9 on 2024-01-23 14:24

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("caluma_workflow", "0031_simple_history"),
    ]

    operations = [
        migrations.AlterField(
            model_name="workflow",
            name="start_tasks",
            field=models.ManyToManyField(
                help_text="Starting task(s) of the workflow.",
                related_name="+",
                to="caluma_workflow.task",
            ),
        ),
    ]



# Generated by Django 4.2.10 on 2024-05-31 06:26

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("caluma_form", "0047_alter_answer_documents"),
    ]

    operations = [
        migrations.AddField(
            model_name="historicaloption",
            name="is_hidden",
            field=models.TextField(default="false"),
        ),
        migrations.AddField(
            model_name="option",
            name="is_hidden",
            field=models.TextField(default="false"),
        ),
    ]



from django.conf import settings


class BaseUser:  # pragma: no cover
    def __init__(self, username=None, groups=None, group=None, token=None, claims=None):
        self.username = username
        self.groups = [] if groups is None else groups
        self.group = group
        self.token = token
        self.claims = {} if claims is None else claims
        self.is_authenticated = False

    def __str__(self):
        raise NotImplementedError


class AnonymousUser(BaseUser):
    def __str__(self):
        return "AnonymousUser"


class OIDCUser(BaseUser):
    def __init__(self, token: str, claims: dict):
        super().__init__(token=token, claims=claims)

        self.username = self.claims[settings.OIDC_USERNAME_CLAIM]
        self.groups = self.claims.get(settings.OIDC_GROUPS_CLAIM, [])
        self.group = self.groups[0] if self.groups else None
        self.token = token
        self.is_authenticated = True

    def __str__(self):
        return self.username



# Generated by Django 4.2.15 on 2024-12-02 13:43

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('caluma_workflow', '0033_workitem_indexes'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='workitem',
            index=models.Index(fields=['created_at'], name='caluma_work_created_c87fd3_idx'),
        ),
        migrations.AddIndex(
            model_name='workitem',
            index=models.Index(fields=['deadline'], name='caluma_work_deadlin_3d2f3c_idx'),
        ),
    ]



from django.core.management.base import BaseCommand

from caluma.caluma_form.models import Answer, Question
from caluma.caluma_form.utils import update_or_create_calc_answer


class Command(BaseCommand):
    """
    Recalculate calculated answers containing values from TableAnswers.

    Due to a bug, answers to calculated questions were wrong, if they contained values
    from table rows. This command recalculates all of them.
    """

    help = "Recalculate calculated answers containing values from TableAnswers."

    def handle(self, *args, **options):
        affected_questions = (
            Question.objects.filter(type="table")
            .exclude(calc_dependents=[])
            .values_list("calc_dependents", flat=True)
        )

        affected_questions_clean = set(
            [slug for entry in affected_questions for slug in entry]
        )

        calc_answers = Answer.objects.filter(
            question__type="calculated_float",
            document__isnull=False,
            question__slug__in=affected_questions_clean,
        )

        for answer in calc_answers.iterator():
            update_or_create_calc_answer(answer.question, answer.document)



# Generated by Django 4.2.16 on 2024-09-26 08:21

import django.contrib.postgres.indexes
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('caluma_workflow', '0032_alter_workflow_start_tasks'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='workitem',
            index=models.Index(fields=['status'], name='caluma_work_status_4d2ecc_idx'),
        ),
        migrations.AddIndex(
            model_name='workitem',
            index=django.contrib.postgres.indexes.GinIndex(fields=['controlling_groups'], name='caluma_work_control_6ede39_gin'),
        ),
    ]



"""
Structure - Fast and correct form structure representation.

Design requirements:

* Fast initialisation from a document. Use preload if needed to reduce
  number of queries
* Eager loading of a full document structure. No lazy-loading with questionable
  performance expectations
* Correct navigation via question slugs (JEXL references) from any context
* Fast lookups once initialized
* Extensible for caching JEXL expression results

* No properties. Code is always in methods
"""

from __future__ import annotations

import collections
import copy
import itertools
import typing
import weakref
from abc import ABC
from collections import defaultdict
from collections.abc import Iterable
from dataclasses import dataclass, field
from functools import singledispatch, wraps
from logging import getLogger
from typing import Optional

from caluma.caluma_core import exceptions

if typing.TYPE_CHECKING:  # pragma: no cover
    from caluma.caluma_form.jexl import QuestionJexl

from caluma.caluma_form.models import (
    Answer,
    AnswerDocument,
    Document,
    DynamicOption,
    File,
    Form,
    FormQuestion,
    Option,
    Question,
    QuestionOption,
)

log = getLogger(__name__)


def object_local_memoise(method):
    """Decorate a method to become object-local memoised.

    In other words - The method will cache it's results. If the method is called
    twice with the same arguments, it will return the cached result instead.

    For debugging purposes, you can also set `object_local_memoise.enabled`
    to `False`, which will then behave just as if the memoising didn't happen.
    """

    @wraps(method)
    def new_method(self, *args, **kwargs):
        if not object_local_memoise.enabled:  # pragma: no cover
            # for debugging purposes
            return method(self, *args, **kwargs)
        if not hasattr(self, "_memoise"):
            clear_memoise(self)

        key = str([args, kwargs, method])
        if key in self._memoise:
            object_local_memoise.hit_count += 1
            self._memoise_hit_count += 1
            return self._memoise[key]
        ret = method(self, *args, **kwargs)
        self._memoise_miss_count += 1
        object_local_memoise.miss_count += 1
        self._memoise[key] = ret
        return ret

    return new_method


# This should only be set to `False` for debugging
setattr(object_local_memoise, "enabled", True)

# Statistics - for analysis / debugging
setattr(object_local_memoise, "hit_count", 0)
setattr(object_local_memoise, "miss_count", 0)


def clear_memoise(obj):
    """Clear memoise cache for given object.

    If an object uses the `@object_local_memoise` decorator, you can then
    call `clear_memoise()` on that object to clear all it's cached data.
    """
    obj._memoise = {}
    obj._memoise_hit_count = 0
    obj._memoise_miss_count = 0


class FastLoader:
    """Load everything in a document/form combination as fast as possible.

    The FastLoader is intended to reduce the number of queries as much as
    possible. It uses prefetching and other techniques, and provides fast (dict)
    lookup to any involved entities.

    Note: We treat all keys as strings, even UUIDs. This makes us a bit
    more compatible with various sources of keys.

    While you *can* use the fastloader with any document, it will likely
    load too much if only a sub-document is given
    """

    @classmethod
    def for_document(cls, document):
        new_self = cls()
        new_self._load_document_entities([document.family])
        new_self._load_form_entities(
            set(doc.form_id for doc in new_self._documents.values())
        )
        return new_self

    @classmethod
    def for_queryset(cls, document_qs):
        new_self = cls()
        new_self._load_document_entities(document_qs)
        new_self._load_form_entities(
            set(doc.form_id for doc in new_self._documents.values())
        )
        return new_self

    def __init__(self):
        # Lazy import needed here
        from caluma.caluma_form.jexl import QuestionJexl

        # Evaluator is only used for extracting answer transforms & friends,
        # so is field-independent here
        self._evaluator = QuestionJexl(field=None)

        # Form bits
        self._forms: dict[str, Form] = {}
        self._questions: dict[str, Question] = {}
        self._questions_by_form: dict[str, list[Question]] = defaultdict(list)
        self._question_options: dict[str, list[Option]] = defaultdict(list)
        self._answers_by_document = defaultdict(dict)

        # Document bits
        self._documents: dict[str, Document] = {}
        self._answers: dict[str, Answer] = {}
        self._table_rows_by_answer = defaultdict(list)
        self._answerdocuments: dict[str, AnswerDocument] = {}
        self._files: dict[str, File] = {}
        self._files_by_answer = defaultdict(list)
        self._dynamic_options_by_question = defaultdict(dict)

        # Jexl dependency graph: question(A) -> question(B) -> list of expr type
        # Question B depends on Question A via expression "type". May be multiple types
        # This is "reversed" due to the fact that we mostly need "downstream"
        # updates when updating (recalculating) a field
        self._jexl_dependencies: dict[str, dict[str, str]] = defaultdict(
            lambda: defaultdict(list)
        )

    def _store_question(self, question):
        self._questions[question.pk] = question

        # Build a comprehensive dependency graph for all JEXL expressions
        for expr_property in ["is_hidden", "calc_expression", "is_required"]:
            jexl_expr = getattr(question, expr_property)
            if not jexl_expr or "|" not in str(jexl_expr):
                # not interesting - no transform in the expr
                continue
            # Evaluator is only used for extracting answer transforms & friends,
            # so is field-independent here
            referenced_questions = set(
                itertools.chain(
                    self._evaluator.extract_referenced_questions(jexl_expr),
                    self._evaluator.extract_referenced_mapby_questions(jexl_expr),
                )
            )
            for dependency_slug in referenced_questions:
                self._jexl_dependencies[dependency_slug][question.pk].append(
                    expr_property
                )

    def dependents_of_question(self, slug):
        """Return the dependents of the given question.

        Return format is a dict of question slug -> list-of-dependency-type
        for the given question.

        For example: `question-a` mentions `question-b` in it's `is_hidden`
        expression, then the following call would hold true:

        >>> dependents_of_question('question-b')
        {'question-a': ['is_hidden']}
        """
        return self._jexl_dependencies[slug]

    def _load_form_entities(self, known_forms):
        form_questions = (
            FormQuestion.objects.filter(form__in=known_forms)
            .select_related("question")
            .select_related("form")
            .order_by("-sort")
        )
        choice_questions = []

        collected_forms = set()

        for fq in form_questions:
            if fq.form_id not in self._forms:
                self._forms[fq.form_id] = fq.form

            # This is already pre-sorted, so we can naively append() here
            self._questions_by_form[fq.form_id].append(fq.question)
            self._store_question(fq.question)

            if fq.question.type == Question.TYPE_TABLE:
                collected_forms.add(fq.question.row_form_id)
            if fq.question.type == Question.TYPE_FORM:
                collected_forms.add(fq.question.sub_form_id)

            # Prepare next step - fetch ordered choices
            if fq.question.type in [
                Question.TYPE_CHOICE,
                Question.TYPE_MULTIPLE_CHOICE,
            ]:
                choice_questions.append(fq.question.pk)

        # It could be that the requested forms don't have any questions. We'd still
        # need to add them to our "known" set.
        # TODO: Maybe we can just set *some* value in self._forms, if nobody actually
        # wants to *use* the form objects themselves
        missing_forms = set(known_forms) - set(self._forms)
        if missing_forms:
            for form in Form.objects.filter(slug__in=missing_forms):
                self._forms[form.pk] = form

        if choice_questions:
            already_have_options = set(self._question_options.keys())
            newly_needed = set(choice_questions) - already_have_options
            for qo in (
                QuestionOption.objects.filter(question__in=newly_needed)
                .order_by("-sort")
                .select_related("option")
            ):
                self._question_options[qo.question_id].append(qo.option)

        newly_collected_forms = collected_forms - set(self._forms.keys())
        if newly_collected_forms:
            self._load_form_entities(list(newly_collected_forms))

    def _load_document_entities(self, documents):
        # First: All Documents - These are fetchable via zero JOINs
        if isinstance(documents, list):
            families = [doc.family_id for doc in documents]
        else:
            families = documents.values("family_id")

        self._documents = {
            str(document.pk): document
            for document in Document.objects.filter(family__in=families)
        }
        # Second: All Answers - These are fetchable via zero JOINs, as we
        # already have all the documents
        self._load_answers(self._documents.keys())

        self._answerdocuments = {
            str(answerdocument.pk): answerdocument
            for answerdocument in AnswerDocument.objects.filter(
                # This works, as there should only be the root doc, and any
                # other document in here will be a row doc. So one mismatch,
                # the rest should match just fine
                document__in=self._documents.keys()
            ).order_by("answer_id", "-sort")
        }
        for ad in self._answerdocuments.values():
            self._table_rows_by_answer[str(ad.answer_id)].append(str(ad.document_id))

        # TODO: This uses a JOIN despite us having the answers already loaded.
        # However we don't know the answer's question types yet, so instead of passing
        # a big number of answer ids here (only a few of which will be file answers)
        # we instead use the documents, reducing the number of params, and therefore
        # speeding up the query.
        for file in File.objects.filter(
            answer__document__in=self._documents.keys()
        ).order_by("answer_id", "created_at"):
            self._files_by_answer[str(file.answer_id)].append(file)

        self._load_dynamic_options()

    def _load_answers(self, documents):
        new_answers = {
            str(answer.pk): answer
            for answer in Answer.objects.filter(document__in=documents)
        }
        self._answers.update(new_answers)
        for ans in new_answers.values():
            self._answers_by_document[str(ans.document_id)][ans.question_id] = ans

    def _load_dynamic_options(self):
        for do in DynamicOption.objects.filter(document__in=self._documents):
            self._dynamic_options_by_question[do.question_id][do.slug] = do

    def question_for_answer(self, answer_id):
        ans = self._answers[str(answer_id)]
        return self._questions[ans.question_id]

    def answers_for_document(self, document_id: str) -> dict[str, Answer]:
        """Return an unordered dict of all answers in the given document.

        The resulting dict is keyed by the question slug, pointing to the
        answer object.
        """
        return self._answers_by_document[str(document_id)]

    def files_for_answer(self, answer_id: str) -> list[File]:
        return self._files_by_answer[str(answer_id)]

    def questions_for_form(self, form_id) -> list[Question]:
        return self._questions_by_form[form_id]

    def form_by_id(self, form_id: str) -> Form:
        if form_id not in self._forms:
            log.warning("Fastloader: Form %s was not preloaded - loading now", form_id)
            self._load_form_entities([form_id])
        return self._forms[form_id]

    def rows_for_table_answer(self, answer_id: str) -> list[Document]:
        document_ids = self._table_rows_by_answer[str(str(answer_id))]
        return [self._documents[doc_id] for doc_id in document_ids]

    def options_for_question(self, question_id):
        return self._question_options[question_id]

    def dynamic_options_for_question(self, question_id):
        return self._dynamic_options_by_question[question_id]


@dataclass
class BaseField(ABC):
    """Base class for the field types. This is the interface we aim to provide."""

    parent: Optional["FieldSet"] = field(default=None)

    question: Optional[Question] = field(default=None)
    answer: Optional[Answer] = field(default=None)

    _fastloader: Optional[FastLoader] = field(default=None)

    def _make_fastloader(self, document):
        return FastLoader.for_document(document)

    @object_local_memoise
    def get_evaluator(self) -> QuestionJexl:
        """Return the JEXL evaluator for this field."""

        # JEXL is implemented such that it has one context in the engine, and
        # trying to swap it out to switch context is just problematic. So we use
        # one JEXL instance per field.

        # deferred import to avoid circular dependency
        from caluma.caluma_form.jexl import QuestionJexl

        # The "info" block is provided by the global context, but we need
        # to patch it to represent some local information:  The "form" and
        # "document" bits should point to the current context, not to the global
        # structure. This is a bit unfortunate, but we *have* to fill this in
        # two separate places to avoid breaking compatibility
        context = collections.ChainMap(
            # We need a deep copy of the global context, so we can
            # extend the info block without leaking
            copy.deepcopy(self.get_global_context()),
            self.get_context(),
        )

        context["info"].update(self.get_local_info_context())

        # Legacy form ref - pointing to the root form. Thinking... can we
        # deprecate this in some reasonable way and let users know without
        # breaking stuff?
        context["form"] = context["info"]["root"]["form"]

        return QuestionJexl(field=self, context=context)

    def get_root(self):
        return self.parent.get_root() if self.parent else self

    @object_local_memoise
    def get_local_info_context(self):
        """Return the dictionary to be used in the local `info` context block.

         Properties (See Ember-Caluma's field source for reference):
         - `form`: Legacy property pointing to the root form.
           -> defined in get_evaluator() as we're only returning `info` here
        * Form information
             - `info.form`: The form this question is attached to.
             - `info.formMeta`: The meta of the form this question is attached to.
             - `info.parent.form`: The parent form if applicable.
             - `info.parent.formMeta`: The parent form meta if applicable.
             - `info.root.form`: The new property for the root form.
             - `info.root.formMeta`: The new property for the root form meta.
        * Case information is taken from the global context
             - `info.case.form`: The cases' form (works for task forms and case forms).
             - `info.case.workflow`: The cases' workflow (works for task forms and case forms).
             - `info.case.root.form`: The _root_ cases' form (works for task forms and case forms).
             - `info.case.root.workflow`: The _root_ cases' workflow (works for task forms and case forms).

        """
        form = self.get_form()

        if parent_info := self.get_parent_fieldset():
            parent_data = {
                "form": parent_info.get_form().slug,
                "formMeta": parent_info.get_form().meta,
            }
        else:
            parent_data = None
        return {
            "question": self.question.slug if self.question else None,
            "form": form and form.slug or None,
            "formMeta": form and form.meta or None,
            "parent": parent_data,
            # TODO how is "root" expected to behave if we're *already* on root?
            "root": self.get_root().get_local_info_context() if self.parent else None,
        }

    def get_parent_fieldset(self):
        """Return the parent fieldset, according to JEXL semantics.

        In JEXL, the parent refers to the next field up that represents another
        form. In Rows for example, this is two levels up, but in regular nested
        forms, it's only one level up.
        """
        my_form = self.get_form()
        p = self.parent
        while p and p.get_form() == my_form:
            p = p.parent
        return p

    @object_local_memoise
    def is_required(self) -> bool:
        """Return True if the field is required.

        Evaluate the `is_required` expression of the field. But if the field
        is hidden, it is considered not-required regardless of what the JEXL
        expression says.
        """

        if self.is_hidden():
            # hidden can never be required
            return False
        if self.all_dependencies_hidden_or_empty(self.question.is_required):
            # All dependencies are hidden, which means we can't calculate
            # the expression at all - assume the field is not required regardless
            # of the expression
            return False

        return self.evaluate_jexl(self.question.is_required)

    @object_local_memoise
    def is_visible(self) -> bool:
        """Just return the opposite of is_hidden() for convenience."""
        return not self.is_hidden()

    @object_local_memoise
    def all_dependencies_hidden_or_empty(self, expr):
        """Check if all dependencies of the expression are hidden or empty.

        Note: If there are no dependencies, we return False. Question slugs
        referring to missing fields are ignored.
        """
        dependencies = list(self.get_evaluator().extract_referenced_questions(expr))
        dep_fields = {
            dep: self.get_field(dep) for dep in dependencies if self.get_field(dep)
        }

        return dep_fields and all(
            # Field missing may be acceptable if it's used in an `answer`
            # transform with a default. Therefore if a field is missing here,
            # we consider it as "empty"/"hidden"
            (field.is_hidden() or field.is_empty()) if field else True
            for field in dep_fields.values()
        )

    @object_local_memoise
    def is_hidden(self, raise_on_error=True) -> bool:
        """Return True if the field is hidden.

        A field is hidden if either it's parent is hidden, or it's `is_hidden`
        JEXL expression evaluates to `True`.
        """
        if self.parent and self.parent.is_hidden():
            # If the parent is hidden, then *we* are implicitly also hidden,
            # without even evaluating our own is_hidden expression
            return True

        if not self.question and not self.parent:
            # Root field is always visible
            return False

        if self.all_dependencies_hidden_or_empty(self.question.is_hidden):
            # All dependencies are hidden, which means we can't calculate
            # the expression at all, and therefore we don't show this field
            # in the form.
            return True

        try:
            return self.evaluate_jexl(self.question.is_hidden, raise_on_error)
        except exceptions.QuestionMissing:
            if raise_on_error:
                # if raise_on_error is False, we ignore this one as well.
                # should be used internally only, for example in get_value()
                raise
            # If the expression fails, we assume an error
            # and consider ourselves as hidden
            return True

    @object_local_memoise
    def slug(self):
        return self.question and self.question.slug or None

    def get_context(self) -> collections.ChainMap: ...

    @object_local_memoise
    def get_field(self, slug) -> BaseField:
        return self.get_context().get(slug)

    @object_local_memoise
    def find_field_by_answer(self, answer) -> BaseField:
        q_field = self.get_field(answer.question_id)
        if q_field and q_field.answer and q_field.answer.pk == answer.pk:
            return q_field

        # answer is not in "our" document, probably we're in a row doc.
        # Therefore, search "everywhere"
        for fld in self.get_root().find_all_fields_by_slug(answer.question_id):
            if fld.answer and fld.answer.pk == answer.pk:
                return fld
        return None

    def refresh(self, answer=None):
        """Refresh this field's answer.

        If an answer is given, use it and update the structure in-place.
        Otherwise, look in the DB.

        Also clear out all the caches on our own field as well as all
        the calc dependents.

        Note: Saving the answer is the caller's responsibility, we only
        update the structure in-memory.
        """
        if answer:
            self.answer = answer
        elif self.answer:
            self.answer.refresh_from_db()
        elif not isinstance(self, FieldSet):
            self.answer = Answer.objects.filter(
                question=self.question, document=self.parent._document
            ).first()

        clear_memoise(self)
        for dep in self._fastloader.dependents_of_question(self.slug()):
            for dep_field in self.get_root().find_all_fields_by_slug(dep):
                dep_field.refresh()

    def calculate(self):
        try:
            return self.evaluate_jexl(self.question.calc_expression)
        except Exception:
            # In calc, if an expression evaluation fails, we just return None.
            # We could be in an unattached table row, for example, and we'd
            # do the recalculation when we're actually being attached.
            return None

    def evaluate_jexl(self, expression: str, raise_on_error=True):
        # Some stupid shortcuts to avoid building up an evaluation context for
        # ~90% of cases where the expression is a simple "true" or "false"
        fast_results = {"true": True, "false": False}
        if (fast_result := fast_results.get(expression)) is not None:
            return fast_result

        eval = self.get_evaluator()

        try:
            return eval.evaluate(expression, raise_on_error)
        except exceptions.QuestionMissing:
            raise
        except Exception as exc:
            log.error(
                f"Error while evaluating expression on question {self.slug()}: "
                f"{expression!r}: {str(exc)}"
            )
            if raise_on_error:
                raise RuntimeError(
                    f"Error while evaluating expression on question {self.slug()}: "
                    f"{expression!r}. The system log contains more information"
                )
            # return None is implied

    def get_global_context(self) -> dict:
        return self._global_context

    @object_local_memoise
    def get_form(self):
        """Return the form in which this field *resides*.

        Note this may be different from the .form property, which for
        fieldsets and rowsets will contain the sub/row form instead.
        """
        parent_form = self.parent.form if self.parent else None
        return parent_form or self.form or None

    def get_path(self):  # pragma: no cover
        """Return a path to the current field.

        The path is a human-readable path to the current field, starting from
        the root. This allows having more context visible
        """

        # This is mainly for debugging
        if not self.parent:
            return "(root)"
        ppath = self.parent.get_path()
        typeinfo = type(self).__name__[0]
        return f"{ppath} -> {typeinfo}:{self.slug()}"


@dataclass
class ValueField(BaseField):
    """Represent a field in the form.

    This is roughly 1:1 with a question, but repeated in case of a table
    row for example.
    """

    # empty string is acceptable
    EMPTY_VALUES = ([], (), {}, None)

    @object_local_memoise
    def get_value(self):
        if self.answer and not self.is_hidden(raise_on_error=False):
            if self.answer.value not in self.EMPTY_VALUES:
                return self.answer.value
            if self.question.type == Question.TYPE_DATE and self.answer.date:
                return self.answer.date
            if self.question.type == Question.TYPE_FILES:
                file_objs = self._fastloader.files_for_answer(self.answer.pk)
                return [f.name for f in file_objs]
        return self.question.empty_value()

    @object_local_memoise
    def get_context(self) -> collections.ChainMap:
        return self.parent.get_context()

    def get_options(self):
        return self._fastloader.options_for_question(self.slug())

    def get_dynamic_options(self):
        """Return the dynamic options for this value.

        Note: These are the dynamic options only, not the full data source
        contents!

        Return a dict of the form slug -> option-object:w

        """
        return self._fastloader.dynamic_options_for_question(self.slug())

    @object_local_memoise
    def is_empty(self):
        if not self.answer:
            return True
        is_date = self.question.type == Question.TYPE_DATE
        is_files = self.question.type == Question.TYPE_FILES
        return (
            # Yeah it's weird - get_value should return empty string if that's
            # what the answer has stored, but is_empty() should treat it as
            # empty, still ¯\_(ツ)_/¯
            not bool(
                (self.answer.value not in (*self.EMPTY_VALUES, ""))
                or (is_date and self.answer.date is not None)
                or (is_files and self.answer.files.exists())
            )
            # Being hidden makes you empty even if an answer exists
            or self.is_hidden()
        )

    def get_global_context(self) -> dict:
        return self.parent.get_global_context()

    def __str__(self):
        return f"Field({self.question.slug}, {self.get_value()})"

    def __repr__(self):
        return f"ValueField(q={self.question.slug}, v={self.get_value()})"


class FieldSet(BaseField):
    def __init__(
        self,
        document,
        form=None,
        parent=None,
        question=None,
        global_context=None,
        _fastloader=None,
    ):
        # TODO: prefetch document once we have the structure built up
        #
        self.question = question
        self._document = document

        self._fastloader = _fastloader or self._make_fastloader(document)
        self.form = form or self._fastloader.form_by_id(document.form_id)

        self._global_context = global_context or {"info": {}}

        # uplinks always weak
        self.parent = weakref.proxy(parent) if parent else None

        self._own_fields = {}

        if parent:
            # Our context is an extension of the parent's context. That way, we can
            # see the fields from the parent context.
            self._context = collections.ChainMap(
                self._own_fields, self.parent.get_context()
            )

            # Extending parent's context with our own: This makes our own fields
            # visible in the parent context as well
            self._extend_root_context(self._own_fields)
        else:
            # Root fieldset
            self._context = collections.ChainMap(self._own_fields)

        self._build_context()

    def _extend_root_context(self, new_map):
        if self.parent:
            self.parent._extend_root_context(new_map)
        else:
            self._context.maps.append(new_map)

    @object_local_memoise
    def is_empty(self):
        # Fieldset: If *any* field is non-empty, we consider ourselves also
        # non-empty.
        # We reverse the lookup here to speed up (don't need to see if *all* are
        # empty, just if *one* has a value)
        if self.is_hidden():
            # Hidden makes us empty, even if there's theoretically a value.
            # We do the "cheap" check first, then iterate over the children.
            return True
        has_at_least_one_value = any(not child.is_empty() for child in self.children())
        return not has_at_least_one_value

    def get_context(self):
        return self._context

    def get_value(self):
        if self.is_empty():
            # is_empty() will return True if we're hidden, so
            # no need to double-check
            return {}
        return {
            formfield.question.slug: formfield.get_value()
            for formfield in self._own_fields.values()
        }

    def is_required(self) -> bool:
        # Fieldsets (in other words - subforms) should never be required,
        # regardless of what their JEXL says
        return False

    def get_all_fields(self) -> Iterable[BaseField]:
        """Return all fields in the structure, as an iterator.

        Yields (slug,field) tuples. But note that a slug may be repeated
        as we iterate over table rows.

        NOTE: For tables, the same *question* may be repeated in each row. This
        is intended and normal behaviour.
        """
        for formfield in self._own_fields.values():
            yield formfield
            if isinstance(formfield, FieldSet):
                yield from formfield.get_all_fields()
            elif isinstance(formfield, RowSet):
                # row sets *must* have fieldsets as children. Let's loop
                # over them here
                for child in formfield.children():
                    yield child
                    yield from child.get_all_fields()

    def find_all_fields_by_slug(self, slug: str) -> list[BaseField]:
        """Return all fields with the given question slug.

        This may return multiple fields, as tables are traversed as well.

        Used for recalculation of calc fields, and other places that may need
        to see all fields for a given question.

        If you need the one field that the `answer` transform would return in
        this context, use `.get_field()` instead.
        """
        result = []
        for formfield in self.get_all_fields():
            if formfield.slug() == slug:
                result.append(formfield)
        return result

    def children(self):
        # This is already sorted, as the context buildup
        # is doing that for us.
        return list(self._own_fields.values())

    def _build_context(self):
        # context inheritance: The ChainMap allows lookups in "parent"
        # contexts, so a row context will be able to look "out". We implement
        # form questions the same way, even though not strictly neccessary

        answers_by_q_slug = self._fastloader.answers_for_document(self._document.pk)

        questions = self._fastloader.questions_for_form(self.form.pk)

        for question in questions:
            if question.type == Question.TYPE_FORM:
                self._context[question.slug] = FieldSet(
                    document=self._document,
                    form=self._fastloader.form_by_id(question.sub_form_id),
                    parent=self,
                    question=question,
                    global_context=self.get_global_context(),
                    _fastloader=self._fastloader,
                )
            elif question.type == Question.TYPE_TABLE:
                self._context[question.slug] = RowSet(
                    question=question,
                    answer=answers_by_q_slug.get(question.slug),
                    parent=self,
                    _fastloader=self._fastloader,
                )
            else:
                # "leaf" question
                self._context[question.slug] = ValueField(
                    question=question,
                    answer=answers_by_q_slug.get(question.slug),
                    parent=self,
                    _fastloader=self._fastloader,
                )

    def __str__(self):
        return f"FieldSet({self.form.slug})"

    def __repr__(self):
        q_slug = self.question.slug if self.question else "(root)"

        return f"FieldSet(q={q_slug}, f={self.form.slug})"

    def print_structure(self, print_fn=None, method=str):  # pragma: no cover
        """Print the structure as a hierarchical list of text.

        You can pass an alternate print function, and also define a custom
        function for turning the fields into text. For example, calling with
        method=repr gives more information than the default method=str.
        """
        return print_structure(self, print_fn, method=method)

    def list_structure(self, method=str):  # pragma: no cover
        """Return the structure as a  list of text, indented.

        You can pass an alternate print function, and also define a custom
        function for turning the fields into text. For example, calling with
        method=repr gives more information than the default method=str.
        """
        return list_structure(self, method)


class RowSet(BaseField):
    rows: list[FieldSet]

    def __init__(
        self, question, parent, answer: Optional[Answer] = None, _fastloader=None
    ):
        self.form = question.row_form
        self.question = question
        self.answer = answer
        self._fastloader = _fastloader or self._make_fastloader(parent._document)

        if not parent:  # pragma: no cover
            raise exceptions.ConfigurationError(
                f"Table question {self.slug()} has no parent"
            )

        self.parent = weakref.proxy(parent)

        if answer:
            self.rows = [
                FieldSet(
                    document=row_doc,
                    question=question,
                    form=self._fastloader.form_by_id(question.row_form_id),
                    parent=self,
                    global_context=self.get_global_context(),
                    _fastloader=self._fastloader,
                )
                for row_doc in self._fastloader.rows_for_table_answer(answer.pk)
            ]
        else:
            self.rows = []

    def get_value(self):
        if self.is_hidden():  # pragma: no cover
            return []

        return [row.get_value() for row in self.children()]

    @object_local_memoise
    def get_column_questions(self):
        return self._fastloader.questions_for_form(self.form.pk)

    def get_all_fields(self) -> Iterable[BaseField]:
        for row in self.children():
            yield row
            # Row field children are always FieldSets
            yield from row.get_all_fields()

    def get_context(self):
        # Rowset does not have it's own context: Any field within is
        # basically in it's own world (but has a view onto the "outside")
        return self.parent.get_context()

    def children(self):
        return self.rows

    def _extend_root_context(self, new_map):
        # Rowset: We do not recurse further up when extending context,
        # and we're also not updating our own context from the rows
        pass

    @object_local_memoise
    def get_global_context(self) -> dict:
        if not self.parent:  # pragma: no cover
            raise exceptions.ConfigurationError(
                f"Table question {self.slug()} has no parent"
            )

        return self.parent.get_global_context()

    @object_local_memoise
    def is_empty(self):
        # Table is considered empty if it has no rows.
        # Hidden implies empty, even if there *theoretically* is an answer
        # present
        return self.is_hidden() or not bool(self.rows)

    def __str__(self):
        return f"RowSet({self.form.slug})"

    def __repr__(self):
        return f"RowSet(q={self.question.slug}, f={self.form.slug})"


def print_structure(fieldset: FieldSet, print_fn=None, method=str):
    """Print a document's structure.

    Intended halfway as an example on how to use the structure
    classes, and a debugging utility.
    """
    ind = {"i": 0}

    print_fn = print_fn or print

    @singledispatch
    def visit(vis):  # pragma: no cover
        # Should never happen - for completeness only
        raise Exception(f"generic visit(): {vis}")

    @visit.register(FieldSet)
    def _(vis: FieldSet):
        print_fn("   " * ind["i"], method(vis))
        ind["i"] += 1
        for sub in vis.children():
            visit(sub)
        ind["i"] -= 1

    @visit.register(RowSet)
    def _(vis: RowSet):
        print_fn("   " * ind["i"], method(vis))
        ind["i"] += 1
        for sub in vis.children():
            visit(sub)
        ind["i"] -= 1

    @visit.register(ValueField)
    def _(vis):
        print_fn("   " * ind["i"], method(vis))

    visit(fieldset)


def list_structure(fieldset, method=str):
    """List the given fieldset's structure.

    Use the given method (Default: str()) for stringification.
    Return a list of strings, each representing a field in the structure,
    properly indented, useful for visualisation
    """
    out_lines = []

    def fake_print(*args):
        out_lines.append(" ".join([x for x in args]))

    print_structure(fieldset, print_fn=fake_print, method=method)
    return out_lines


def list_document_structure(document, method=str):
    """List the given document's structure.

    Use the given method (Default: str()) for stringification.
    """
    fs = FieldSet(document)
    return list_structure(fs, method)



import sys
from datetime import date
from logging import getLogger

from django_filters.constants import EMPTY_VALUES
from rest_framework import exceptions

from caluma.caluma_core.exceptions import ConfigurationError
from caluma.caluma_data_source.data_source_handlers import get_data_sources
from caluma.caluma_form import structure
from caluma.caluma_workflow.models import Case

from . import jexl, models
from .format_validators import get_format_validators
from .models import DynamicOption, Question

log = getLogger()


class CustomValidationError(exceptions.ValidationError):
    """Custom validation error to carry more information.

    This can carry more information about the error, so the documentValidity
    query can show more useful information.
    """

    def __init__(self, detail, code=None, slugs=None):
        slugs = slugs if slugs else []
        super().__init__(detail, code)
        self.slugs = slugs


class AnswerValidator:
    def _validate_question_text(self, question, value, **kwargs):
        min_length = question.min_length if question.min_length is not None else 0
        max_length = (
            question.max_length if question.max_length is not None else sys.maxsize
        )

        if (
            not isinstance(value, str)
            or len(value) < min_length
            or len(value) > max_length
        ):
            raise CustomValidationError(
                f"Invalid value {value}. "
                f"Should be of type str, and it's length between {min_length} and {max_length}",
                slugs=[question.slug],
            )

    def _validate_question_textarea(self, question, value, **kwargs):
        self._validate_question_text(question, value)

    def _validate_question_float(self, question, value, **kwargs):
        min_value = (
            question.min_value if question.min_value is not None else float("-inf")
        )
        max_value = (
            question.max_value if question.max_value is not None else float("inf")
        )

        if not isinstance(value, float) or value < min_value or value > max_value:
            raise CustomValidationError(
                f"Invalid value {value}. "
                f"Should be of type float, not lower than {min_value} "
                f"and not greater than {max_value}",
                slugs=[question.slug],
            )

    def _validate_question_integer(self, question, value, **kwargs):
        min_value = (
            question.min_value if question.min_value is not None else float("-inf")
        )
        max_value = (
            question.max_value if question.max_value is not None else float("inf")
        )

        if not isinstance(value, int) or value < min_value or value > max_value:
            raise CustomValidationError(
                f"Invalid value {value}. "
                f"Should be of type int, not lower than {min_value} "
                f"and not greater than {max_value}",
                slugs=[question.slug],
            )

    def _validate_question_date(self, question, value, **kwargs):
        if not isinstance(value, date):
            raise CustomValidationError(
                f"Invalid value {value}. Should be of type date.", slugs=[question.slug]
            )

    def _structure_field(self, document, question, validation_context):
        """Return the requested structure field, and the root context.

        The root context is mostly not used, but needed to ensure it won't go
        out of scope before the field (form structure has only weakrefs from
        child to parent, and we don't want the parent to go out of scope, as it
        will break child (field) behaviour).
        """
        if validation_context:  # pragma: todo cover
            # Note about coverage: This is in fact covered, but in some python
            # versions, pytest-cov fails to see it when run with xdist, and I
            # don't have the patience to actually debug that stuff right now
            # The test that covers this branch is here:
            # `caluma_form.tests.test_option.test_validate_form_with_jexl_option`

            # If valdiation context is passed in from the DocumentValidator, we
            # already have the *field* we need, and no further work is needed.
            if validation_context.slug() != question.slug:  # pragma: no cover
                # This only happens if *programmer* made an error, therefore we're
                # not explicitly covering it
                raise ConfigurationError(
                    f"Passed validation context does not belong to question {question.slug}"
                )
            return validation_context, validation_context.get_root()

        # If we need to create the context ourselves here, we'll need to fetch
        # the field from the context.
        validation_context = DocumentValidator().get_validation_context(document)
        return validation_context.get_field(question.slug), validation_context

    def _evaluate_options_jexl(
        self, document, question, validation_context=None, qs=None
    ):
        """Return a list of slugs that are, according to their is_hidden JEXL, visible."""

        if not document and not validation_context:
            # Saving options of a default answer. Can't evaluate the options'
            # JEXL anyway, so allow all of them (and hit the DB, don't optimize
            # and use the structure code)
            return [o.slug for o in question.options.all()]

        field, _root = self._structure_field(document, question, validation_context)
        if not field:  # pragma: no cover
            # This only happens if *programmer* made an error, therefore we're
            # not explicitly covering it
            raise ConfigurationError(
                f"Field for question '{question.slug}' not found "
                f"in form '{document.form_id}'"
            )

        options = field.get_options()
        return [o.slug for o in options if not field.evaluate_jexl(o.is_hidden)]

    def visible_options(self, document, question, qs):
        return self._evaluate_options_jexl(document, question, qs=qs)

    def _validate_question_choice(
        self, question, value, validation_context, document, **kwargs
    ):
        options = self._evaluate_options_jexl(
            document, question, validation_context=validation_context
        )

        if not isinstance(value, str) or value not in options:
            raise CustomValidationError(
                f"Invalid value {value}. "
                f"Should be of type str and one of the options {', '.join(options)}",
                slugs=[question.slug],
            )

    def _validate_question_multiple_choice(
        self, question, value, validation_context, document, **kwargs
    ):
        options = self._evaluate_options_jexl(
            document, question, validation_context=validation_context
        )

        invalid_options = set(value) - set(options)
        if not isinstance(value, list) or invalid_options:
            raise CustomValidationError(
                f"Invalid options [{', '.join(invalid_options)}]. "
                f"Should be one of the options [{', '.join(options)}]",
                slugs=[question.slug],
            )

    def _validate_question_dynamic_choice(
        self, question, value, document, user, data_source_context, **kwargs
    ):
        if not isinstance(value, str):
            raise CustomValidationError(
                f'Invalid value "{value}". Must be of type str.', slugs=[question.slug]
            )

        self._validate_dynamic_option(
            question, document, value, user, data_source_context
        )
        self._remove_unused_dynamic_options(question, document, [value])

    def _validate_question_dynamic_multiple_choice(
        self, question, value, document, user, data_source_context, **kwargs
    ):
        if not isinstance(value, list):
            raise CustomValidationError(
                f'Invalid value: "{value}". Must be of type list', slugs=[question.slug]
            )

        for v in value:
            if not isinstance(v, str):
                raise CustomValidationError(
                    f'Invalid value: "{v}". Must be of type string',
                    slugs=[question.slug],
                )
            self._validate_dynamic_option(
                question, document, v, user, data_source_context
            )

        self._remove_unused_dynamic_options(question, document, value)

    def _validate_dynamic_option(
        self, question, document, option, user, data_source_context
    ):
        data_source = get_data_sources(dic=True)[question.data_source]
        data_source_object = data_source()

        valid_label = data_source_object.validate_answer_value(
            option, document, question, user, data_source_context
        )
        if valid_label is False:
            raise CustomValidationError(
                f'Invalid value "{option}". Not a valid option.', slugs=[question.slug]
            )

        DynamicOption.objects.get_or_create(
            document=document,
            question=question,
            slug=option,
            defaults={
                "label": valid_label,
                "created_by_user": user.username,
                "created_by_group": user.group,
            },
        )

    def _remove_unused_dynamic_options(self, question, document, used_values):
        DynamicOption.objects.filter(document=document, question=question).exclude(
            slug__in=used_values
        ).delete()

    def _validate_question_table(
        self,
        question,
        value,
        document,
        user,
        validation_context: structure.RowSet,
        instance=None,
        origin=False,
        **kwargs,
    ):
        if not origin:
            # Rowsets should always be validated in context, so we can use that
            for row in validation_context.children():
                DocumentValidator().validate(
                    row._document, user=user, **kwargs, validation_context=row
                )
            return

        # this answer was the entry-point of the validation
        # -> validate row form
        value = value or instance and instance.documents.all() or []
        question = question or instance and instance.question

        for row_doc in value:
            if row_doc.form_id != question.row_form_id:
                raise exceptions.ValidationError(
                    f"Document {row_doc.pk} is not of form type {question.row_form.pk}."
                )

    def _validate_question_files(self, question, value, **kwargs):
        if not isinstance(value, list) and value is not None:  # pragma: no cover
            # should already have been rejected by schema / gql level
            raise exceptions.ValidationError("Input files must be a list")

    def _validate_question_calculated_float(self, question, value, **kwargs):
        pass

    def validate(
        self,
        *,
        question,
        document,
        user,
        validation_context=None,
        instance=None,
        origin=False,
        data_source_context=None,
        **kwargs,
    ):
        # Check all possible fields for value
        value = None
        for i in ["documents", "files", "date", "value"]:
            value = kwargs.get(i, value)
            if value:
                break

        # empty values are allowed
        # required check will be done in DocumentValidator
        if value in EMPTY_VALUES:
            return

        validate_mapping = {
            "text": self._validate_question_text,
            "textarea": self._validate_question_textarea,
            "float": self._validate_question_float,
            "integer": self._validate_question_integer,
            "date": self._validate_question_date,
            "choice": self._validate_question_choice,
            "multiple_choice": self._validate_question_multiple_choice,
            "dynamic_choice": self._validate_question_dynamic_choice,
            "dynamic_multiple_choice": self._validate_question_dynamic_multiple_choice,
            "table": self._validate_question_table,
            "files": self._validate_question_files,
            "calculated_float": self._validate_question_calculated_float,
        }

        validate_func = validate_mapping[question.type]
        validate_func(
            question,
            value,
            document=document,
            user=user,
            validation_context=validation_context,
            instance=instance,
            origin=origin,
            data_source_context=data_source_context,
        )

        format_validators = get_format_validators(dic=True)
        for validator_slug in question.format_validators:
            format_validators[validator_slug]().validate(value, document)


class DocumentValidator:
    def validate(
        self,
        document,
        user,
        validation_context=None,
        data_source_context=None,
        **kwargs,
    ):
        if not validation_context:
            validation_context = self.get_validation_context(document)

        required_but_empty = []

        all_fields = list(validation_context.get_all_fields())
        for field in all_fields:
            if field.is_required() and field.is_empty():
                required_but_empty.append(field)

        if required_but_empty:
            # When dealing with tables, the same question slug
            # might be missing multiple times. So we're normalizing a bit here
            affected_slugs = sorted(
                set([f.slug() for f in required_but_empty if f.slug()])
            )

            question_s, is_are = (
                ("Question", "is") if len(affected_slugs) == 1 else ("Questions", "are")
            )
            raise CustomValidationError(
                f"{question_s} {', '.join(affected_slugs)} "
                f"{is_are} required but not provided.",
                slugs=affected_slugs,
            )

        for field in all_fields:
            if field.is_visible() and field.answer:
                # if answer is not given, the required_but_empty check above would
                # already have raised an exception
                validator = AnswerValidator()
                validator.validate(
                    document=field.answer.document,
                    question=field.question,
                    value=field.answer.value,
                    documents=field.answer.documents.all(),
                    user=user,
                    validation_context=field,
                    data_source_context=data_source_context,
                )

    def get_validation_context(self, document, _fastloader=None):
        relevant_case: Case = (
            getattr(document, "case", None)
            or getattr(getattr(document, "work_item", None), "case", None)
            or None
        )

        case_form = (
            relevant_case.document.form_id
            if relevant_case and relevant_case.document
            else None
        )
        case_family_form = (
            relevant_case.family.document.form_id
            if relevant_case and relevant_case.family.document
            else None
        )
        case_info = (
            {
                # Why are we even represent this in context of the case?
                # The form does not belong there, it's not even there in
                # the model
                "form": case_form,
                "workflow": relevant_case.workflow_id,
                "root": {
                    "form": case_family_form,
                    "workflow": relevant_case.family.workflow_id,
                },
            }
            if relevant_case
            else None
        )
        workitem_info = (
            document.work_item.__dict__ if hasattr(document, "work_item") else None
        )
        context = {
            "info": {
                "document": document.family,
                "form": document.form,
                "case": case_info,
                "work_item": workitem_info,
            }
        }

        return structure.FieldSet(
            document,
            global_context=context,
            _fastloader=_fastloader,
        )

    def visible_questions(self, document, validation_context=None) -> list[Question]:
        """Evaluate the visibility of the questions for the given context.

        This evaluates the `is_hidden` expression for each question to decide
        if the question is visible.
        Return a list of question slugs that are visible.

        Note: If you pass in a validation context, it's the caller's
        responsibility to ensure it's the right one
        """
        if validation_context is None:
            validation_context = self.get_validation_context(document)

        return [
            field.question
            for field in validation_context.get_all_fields()
            if field.is_visible()
        ]


class QuestionValidator:
    @staticmethod
    def _validate_format_validators(data):
        format_validators = data.get("format_validators")
        if format_validators:
            fv = get_format_validators(include=format_validators, dic=True)
            diff_list = list(set(format_validators) - set(fv))
            if diff_list:
                raise exceptions.ValidationError(
                    f"Invalid format validators {diff_list}."
                )

    @staticmethod
    def _validate_data_source(data_source):
        data_sources = get_data_sources(dic=True)
        if data_source not in data_sources:
            raise exceptions.ValidationError(f'Invalid data_source: "{data_source}"')

    @staticmethod
    def _validate_calc_expression(data):
        """Don't allow other calc questions or non-existent questions."""
        expr = data.get("calc_expression")
        if not expr:
            return

        question_jexl = jexl.QuestionJexl(field=None)
        deps = set(question_jexl.extract_referenced_questions(expr))

        inexistent_slugs = deps - set(
            models.Question.objects.filter(pk__in=deps).values_list("slug", flat=True)
        )
        illegal_deps = ", ".join(inexistent_slugs)

        if illegal_deps:  # pragma: no cover
            raise exceptions.ValidationError(
                f"Calc expression references inexistent questions: {illegal_deps}"
            )

    def validate(self, data):
        if data["type"] in [models.Question.TYPE_TEXT, models.Question.TYPE_TEXTAREA]:
            self._validate_format_validators(data)
        elif data["type"] == models.Question.TYPE_CALCULATED_FLOAT:
            self._validate_calc_expression(data)

        if "dataSource" in data:
            self._validate_data_source(data["dataSource"])


def get_document_validity(document, user, **kwargs):
    validator = DocumentValidator()
    is_valid = True
    errors = []

    try:
        validator.validate(document, user, **kwargs)
    except CustomValidationError as exc:
        is_valid = False
        detail = str(exc.detail[0])
        errors = [{"slug": slug, "error_msg": detail} for slug in exc.slugs]

    return {"id": document.id, "is_valid": is_valid, "errors": errors}



# Generated by Django 4.2.9 on 2024-01-23 14:24

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("caluma_form", "0046_file_answer_reverse_keys"),
    ]

    operations = [
        migrations.AlterField(
            model_name="answer",
            name="documents",
            field=models.ManyToManyField(
                related_name="+",
                through="caluma_form.AnswerDocument",
                to="caluma_form.document",
            ),
        ),
    ]



"""
Caluma specific settings.

This settings module only contains caluma-internal settings. It's imported
by the main caluma settings and is intended to also be used by third party
applications integrating Caluma.
"""

import os

import environ

env = environ.Env()
django_root = environ.Path(__file__) - 3

ENV_FILE = env.str("ENV_FILE", default=django_root(".env"))
if os.path.exists(ENV_FILE):  # pragma: no cover
    environ.Env.read_env(ENV_FILE)

# per default production is enabled for security reasons
# for development create .env file with ENV=development
ENV = env.str("ENV", "production")


def default(default_dev=env.NOTSET, default_prod=env.NOTSET):
    """Environment aware default."""
    return default_prod if ENV == "production" else default_dev


# Managing files

MEDIA_STORAGE_SERVICE = env.str("MEDIA_STORAGE_SERVICE", default="minio")
MINIO_STORAGE_ENDPOINT = env.str("MINIO_STORAGE_ENDPOINT", default="minio:9000")
MINIO_STORAGE_ACCESS_KEY = env.str("MINIO_STORAGE_ACCESS_KEY", default="minio")
MINIO_STORAGE_SECRET_KEY = env.str("MINIO_STORAGE_SECRET_KEY", default="minio123")
MINIO_STORAGE_USE_HTTPS = env.str("MINIO_STORAGE_USE_HTTPS", default=False)
MINIO_STORAGE_MEDIA_BUCKET_NAME = env.str(
    "MINIO_STORAGE_MEDIA_BUCKET_NAME", default="caluma-media"
)
MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET = env.str(
    "MINIO_STORAGE_AUTO_CREATE_MEDIA_BUCKET", default=True
)
MINIO_PRESIGNED_TTL_MINUTES = env.int("MINIO_PRESIGNED_TTL_MINUTES", default=15)
MINIO_DISABLE_CERT_CHECKS = env.bool("MINIO_DISABLE_CERT_CHECKS", default=False)


# GraphQL

GRAPHENE = {
    "SCHEMA": "caluma.schema.schema",
    "MIDDLEWARE": [],
    "RELAY_CONNECTION_MAX_LIMIT": None,
}

# If you set DISABLE_INTROSPECTION to True, any GQL client will not be able to
# query the types and connections, making crafting queries much harder (but not
# impossible, and this is an Open Source product, so they can still go check
# the source).
DISABLE_INTROSPECTION = env.bool("DISABLE_INTROSPECTION", default=default(False, True))

# DOS protection: Limit query depth to a given level. Default is 0, which means
# it is disabled
QUERY_DEPTH_LIMIT = env.int("QUERY_DEPTH_LIMIT", default=0)

# OpenID connect

OIDC_USERINFO_ENDPOINT = env.str("OIDC_USERINFO_ENDPOINT", default=None)
OIDC_VERIFY_SSL = env.bool("OIDC_VERIFY_SSL", default=True)
OIDC_GROUPS_CLAIM = env.str("OIDC_GROUPS_CLAIM", default="caluma_groups")
OIDC_USERNAME_CLAIM = env.str("OIDC_USERNAME_CLAIM", default="sub")
# Claim to use for the "client" claim. Should normally be left to the default value
OIDC_CLIENT_CLAIM = env.str("OIDC_CLIENT_CLAIM", default="client_id")

OIDC_BEARER_TOKEN_REVALIDATION_TIME = env.int(
    "OIDC_BEARER_TOKEN_REVALIDATION_TIME", default=0
)

CALUMA_OIDC_USER_FACTORY = env.str(
    "CALUMA_OIDC_USER_FACTORY", default="caluma.caluma_user.models.OIDCUser"
)


# Extensions

VISIBILITY_CLASSES = env.list(
    "VISIBILITY_CLASSES", default=default(["caluma.caluma_core.visibilities.Any"])
)

PERMISSION_CLASSES = env.list(
    "PERMISSION_CLASSES", default=default(["caluma.caluma_core.permissions.AllowAny"])
)

VALIDATION_CLASSES = env.list("VALIDATION_CLASSES", default=[])

DATA_SOURCE_CLASSES = env.list("DATA_SOURCE_CLASSES", default=[])

FORMAT_VALIDATOR_CLASSES = env.list("FORMAT_VALIDATOR_CLASSES", default=[])

EVENT_RECEIVER_MODULES = env.list("EVENT_RECEIVER_MODULES", default=[])

DYNAMIC_GROUPS_CLASSES = env.list("DYNAMIC_GROUPS_CLASSES", default=[])

DYNAMIC_TASKS_CLASSES = env.list("DYNAMIC_TASKS_CLASSES", default=[])

# simple history
SIMPLE_HISTORY_HISTORY_ID_USE_UUID = True

# Historical API
ENABLE_HISTORICAL_API = env.bool("ENABLE_HISTORICAL_API", default=False)

# Configure the fields you intend to use in the "meta" fields. This will
# provide corresponding constants in the ordreBy filter, as well as allow
# you to use those fields in the analytics module.
META_FIELDS = env.list("META_FIELDS", default=[])

# enable caluma healthz endpoint
ENABLE_HEALTHZ_ENDPOINT = env.bool("ENABLE_HEALTHZ_ENDPOINT", default=False)



import logging
import sys

import pytest
from django.conf import settings
from django.core import management
from django.db import connection
from psycopg import OperationalError
from watchman.decorators import check as watchman_check_decorator

from caluma.caluma_form import storage_clients


@pytest.fixture
def disable_logs():
    """Disable logs for health checks."""
    urllib3_logger = logging.getLogger("urllib3")
    watchman_logger = logging.getLogger("watchman")

    urllib3_level = urllib3_logger.level
    watchman_disabled = watchman_logger.disabled
    urllib3_logger.setLevel(logging.ERROR)
    watchman_logger.disabled = True

    yield

    urllib3_logger.setLevel(urllib3_level)
    watchman_logger.disabled = watchman_disabled


@pytest.fixture
def track_errors(mocker):
    """Write stacktrace to stderr to enable error checking in tests.

    Patches django-watchman check decorator.
    """

    def track_error_check_decorator(func):
        def wrapped(*args, **kwargs):
            response = watchman_check_decorator(func)(*args, **kwargs)

            if response.get("ok") is False:
                sys.stderr.write(response["stacktrace"])
            elif not response.get("ok"):
                for value in response.values():
                    if value.get("ok") is False:
                        sys.stderr.write(value["stacktrace"])

            return response

        return wrapped

    mocker.patch("watchman.decorators.check", new=track_error_check_decorator)

    yield


@pytest.fixture
def db_broken_connection(transactional_db):
    """DB fixture with broken connection.

    Cap DB connection by trying to re-connect with faulty connection details.
    Changes host settings and restores them after use.
    """
    # overwrite DB host setting and re-connect
    try:
        connection.close()
        prev_settings = settings.DATABASES["default"]["HOST"]
        settings.DATABASES["default"]["HOST"] = "not-db"
        connection.connect()
    except OperationalError:
        # ignore connection error
        pass

    yield

    # restore original DB settings
    connection.close()
    settings.DATABASES["default"]["HOST"] = prev_settings
    connection.connect()


@pytest.fixture
def minio_mock_working(mocker):
    """Provide working minio mock for health checks."""
    return mocker.patch.object(
        storage_clients.client.client, "bucket_exists", return_value=True
    )


@pytest.fixture
def minio_not_configured():
    """Fixture with no connection to minio instance.

    Remove minio client to ensure no connection is available.
    """
    # store previous client
    prev_client = storage_clients.client

    # reset client
    storage_clients.client = None

    yield

    # restore previous client
    storage_clients.client = prev_client


@pytest.fixture
def roll_back_migrations():
    """Roll back database migrations to ensure unapplied migrations exist."""
    # undo applied migrations for app contenttypes
    management.call_command("migrate", "contenttypes", "zero")
    try:
        yield
    finally:
        # re-apply migrations for app contenttypes
        management.call_command("migrate", "contenttypes")



import weakref
from collections import ChainMap
from functools import partial

from pyjexl.analysis import ValidatingAnalyzer

from caluma.caluma_core.exceptions import QuestionMissing

from ..caluma_core.jexl import (
    JEXL,
    ExtractTransformArgumentAnalyzer,
    ExtractTransformSubjectAnalyzer,
)

"""
Form JEXL handling

Design principles:

* The JEXL classes do not deal with context switching between questions.
  Where needed, we ask the corresponding field from the structure to give us
  the necessary information.
* The QuestionJexl class only sets up the "runtime", any context is used from
  the structure code.
* We only deal with the *evaluation*, no transform/extraction is happening here
* Caching is done by the structure code, not here
* JEXL evaluation happens lazily, but the results are cached.
"""


class QuestionValidatingAnalyzer(ValidatingAnalyzer):
    def visit_Transform(self, transform):
        if transform.name == "answer" and not isinstance(transform.subject.value, str):
            yield f"{transform.subject.value} is not a valid question slug."

        yield from super().visit_Transform(transform)


class QuestionJexl(JEXL):
    def __init__(self, field, **kwargs):
        """Initialize QuestionJexl.

        Note: The field *may* be set to `None` if you're intending
        to use the object for expression analysis only (exctract_* methods
        for example)
        """
        super().__init__(**kwargs)
        self.field = weakref.proxy(field) if field else None

        self.add_transform("answer", self.answer_transform)

    def answer_transform(self, question_slug, *args):
        field = self.field.get_field(question_slug)

        def _default_or_empty():
            if len(args):
                return args[0]
            return field.question.empty_value()

        if not field:
            if args:
                return args[0]
            else:
                # No default arg, so we must raise an exception
                raise QuestionMissing(
                    f"Question `{question_slug}` could not be found in form {self.field.get_form()}"
                )

        if field.is_hidden():
            # Hidden fields *always* return the empty value, even if we have
            # a default
            return field.question.empty_value()
        elif field.is_empty():
            # not hidden, but empty
            return _default_or_empty()

        return field.get_value()

    def validate(self, expression, **kwargs):
        return super().validate(expression, QuestionValidatingAnalyzer)

    def extract_referenced_questions(self, expr):
        transforms = ["answer"]
        yield from self.analyze(
            expr, partial(ExtractTransformSubjectAnalyzer, transforms=transforms)
        )

    def extract_referenced_mapby_questions(self, expr):
        transforms = ["mapby"]
        yield from self.analyze(
            expr, partial(ExtractTransformArgumentAnalyzer, transforms=transforms)
        )

    def evaluate(self, expr, raise_on_error=True):
        try:
            return super().evaluate(expr, ChainMap(self.context))
        except (
            TypeError,
            ValueError,
            ZeroDivisionError,
            AttributeError,
        ):
            if raise_on_error:
                raise
            return None



from logging import getLogger

from caluma.caluma_form import models, structure, validators
from caluma.caluma_form.jexl import QuestionJexl

log = getLogger(__name__)


def update_calc_dependents(slug, old_expr, new_expr):
    """Update the calc_dependents lists of our calc *dependencies*.

    The given (old and new) expressions are analyzed to see which
    questions are referenced in "our" calc question. Then, those
    questions' calc_dependents list is updated, such that it correctly
    represents the new situation.

    Example: If our expression newly contains the question `foo`, then the
    `foo` question needs to know about it (we add "our" slug to the `foo`s
    calc dependents)
    """
    jexl = QuestionJexl(field=None)
    old_q = set(
        list(jexl.extract_referenced_questions(old_expr))
        + list(jexl.extract_referenced_mapby_questions(old_expr))
    )
    new_q = set(
        list(jexl.extract_referenced_questions(new_expr))
        + list(jexl.extract_referenced_mapby_questions(new_expr))
    )

    to_add = new_q - old_q
    to_remove = old_q - new_q

    questions = models.Question.objects.filter(pk__in=list(to_add | to_remove))

    for question in questions:
        deps = set(question.calc_dependents)
        if question.slug in to_add:
            deps.add(slug)
        else:
            deps.remove(slug)
        question.calc_dependents = list(deps)
        question.save()


def recalculate_field(
    calc_field: structure.ValueField, update_recursively: bool = True
):
    """Recalculate the given value field and store the new answer.

    If it's not a calculated field, nothing happens.
    """
    if calc_field.question.type != models.Question.TYPE_CALCULATED_FLOAT:
        # Not a calc field - skip
        return  # pragma: no cover

    value = calc_field.calculate()

    answer, _ = models.Answer.objects.update_or_create(
        question=calc_field.question,
        document=calc_field.parent._document,
        defaults={"value": value},
    )
    # also save new answer to structure for reuse
    calc_field.refresh(answer)
    if update_recursively:
        recalculate_dependent_fields(calc_field, update_recursively)


def recalculate_dependent_fields(
    changed_field: structure.ValueField, update_recursively: bool = True
):
    """Update any calculated dependencies of the given field.

    If `update_recursively=False` is passed, no subsequent calc dependencies
    are updated (left to the caller in that case).
    """
    for dep_slug in changed_field.question.calc_dependents:
        dep_field = changed_field.get_field(dep_slug)
        if not dep_field:  # pragma: no cover
            # Calculated field is not in our form structure, which is
            # absolutely valid and OK
            continue
        recalculate_field(dep_field, update_recursively)


def update_or_create_calc_answer(question, document, update_dependents=True):
    """Recalculate all answers in the document after calc dependency change."""

    root = validators.DocumentValidator().get_validation_context(document.family)
    for field in root.find_all_fields_by_slug(question.slug):
        recalculate_field(field, update_recursively=update_dependents)



from functools import wraps
from logging import getLogger

from django.db.models import Model

from caluma.caluma_core.types import Node

log = getLogger(__name__)


def is_iterable_and_no_string(value):
    try:
        iter(value)
        return not isinstance(value, str)
    except TypeError:
        return False


def _field_name(db_field):
    if db_field.is_relation:
        return db_field.db_column or f"{db_field.name}_id"
    return db_field.db_column or db_field.name


def col_type_from_db(field, connection):
    """Return a (type, length) tuple for the given field."""

    sql_extract_type = """
        SELECT data_type, character_maximum_length
        FROM information_schema.columns
        WHERE
            column_name = %s AND
            table_name = %s
    """

    cursor = connection.cursor()
    cursor.execute(sql_extract_type, [_field_name(field), field.model._meta.db_table])
    res = cursor.fetchone()
    return res


def fix_foreign_key_types(apps, connection):
    """Ensure all foreign keys pointing to slug fields are correct type."""

    # I'm really not happy that this code needs to exist. I'ts a sign that
    # we somehow messed up and I couldn't reproduce exactly why this even
    # happened in the first place.  Figuring out why it happened would be
    # good, but since the broken version is out there anyway, we need the
    # fix no matter what.

    fix_sql = """
        ALTER TABLE %s
        ALTER COLUMN %s TYPE %s
    """

    for models in apps.all_models.values():
        for model in models.values():
            slug_fks = [
                field
                for field in model._meta.fields
                if field.is_relation and field.target_field.name == "slug"
            ]

            for field in slug_fks:
                # ok we have a slug foreign key
                fk_params = field.db_parameters(connection)
                target_params = field.target_field.db_parameters(connection)

                # verify django-internal specified type
                assert fk_params["type"] == target_params["type"], (
                    f"Foreign key field {field}: type mismatch with destination in django-internal representation"
                )

                # check if the DB agrees
                fk_dbtype = col_type_from_db(field, connection)
                target_dbtype = col_type_from_db(field.target_field, connection)
                if fk_dbtype != target_dbtype:  # pragma: no cover
                    # seems to be fixed upstream
                    cursor = connection.cursor()
                    type_, length = target_dbtype
                    new_type = f"{type_}({length})"
                    log.warning(f"Correcting data type of {field} to {new_type}")
                    # not using the parametrized syntax here (sql, params) as
                    # we do NOT want escaping here
                    cursor.execute(
                        fix_sql % (model._meta.db_table, _field_name(field), new_type)
                    )


def update_model(model: Model, data: dict) -> Model:
    for key, value in data.items():
        setattr(model, key, value)

    model.save()


def disable_raw(signal_handler):  # pragma: no cover
    """Disable signal handlers during loaddata."""

    @wraps(signal_handler)
    def wrapper(*args, **kwargs):
        if kwargs.get("raw"):
            return
        signal_handler(*args, **kwargs)

    return wrapper


def suppressable_visibility_resolver():
    """Make a resolver that can be suppressed in the visibility layer.

    The visibility layer may cause additional load on the database, so
    in cases where it's not needed, the visibility classes can choose to
    skip filtering where it's not absolutely needed.

    For example, if the visibility of a child case is always dependent
    on the associated work item, you can disable enforcing visibility
    on that relationship.

    Example in the GraphQL schema:

    >>> class Form(FormDjangoObjectType):
    >>>     ...
    >>>     resolve_child_case = suppressable_visibility_resolver()

    Example in the visibility class:

    >>> class MyCustomVisibility(BaseVisibility):
    >>>     @filter_queryset_for(Case)
    >>>     def filter_queryset_for_case(self, node, queryset, info):
    >>>         # do the filtering as usual
    >>>         ...
    >>>     ...
    >>>     suppress_visibilities = [
    >>>         "WorkItem.child_case",
    >>>         ...
    >>>     ]

    With the above setting, when looking up child cases from work items,
    the visibility layer would not be run, and the child case would always
    be shown (if the workitem can be shown, of course).

    You may also define it as a `@property` on the visibility class. Note
    however that it's initialized on app startup, you can't change it during
    runtime and make suppression work in some cases and not in others.
    """

    # Avoid circular imports
    from caluma.caluma_core.visibilities import BaseVisibility

    # TODO: This is currently untested (as in unit test), because
    # the checks are done at startup (where the schema is initialized)
    # and we can't swap out the visibility class during test runs
    # to validate the behavour as we'd usually do.
    class SuppressableResolver:
        def __call__(self, inner_self, *args, **kwargs):
            return getattr(inner_self, self.prop, None)

        @property
        def _bypass_get_queryset(self):  # pragma: no cover
            """Tell Graphene to bypass get_queryset()."""
            # If any visibility class tells us to bypass the
            # queryset of this property, we return True
            return any(
                self.qualname in vis_class().suppress_visibilities
                for vis_class in Node.visibility_classes
            )

        def __repr__(self):  # pragma: no cover
            return f"SuppressableResolver({self.qualname})"

        def __set_name__(self, owner, name):
            # Magic. This is called by Python after setting the attribute on
            # the node class, and we use it to figure out which resolver we're
            # building.
            self.name = owner.__name__
            self.prop = name.replace("resolve_", "")
            self.qualname = f"{self.name}.{self.prop}"
            BaseVisibility._suppressable_visibilities.add(self.qualname)

    return SuppressableResolver()
